/*
 * Gaim-Audacious plugin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
#define GAIM_PLUGINS 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "gtkplugin.h"
#include "util.h"
#include "debug.h"
#include "connection.h"
#include "version.h"
#include <audacious/beepctrl.h>

#define GAIMAUD_PLUGIN_ID	"gaim_audacious"

#define OPT_GAIMAUD 		"/plugins/gaim_audacious"
#define OPT_PROCESS_STATUS	OPT_GAIMAUD "/process_status"
#define OPT_PROCESS_USERINFO	OPT_GAIMAUD "/process_userinfo"
#define OPT_PROCESS_ALIAS	OPT_GAIMAUD "/process_alias"

#define SONG_TOKEN		"%song"

#define aud_debug(fmt, ...)	gaim_debug(GAIM_DEBUG_INFO, "Gaim-Audacious", \
					fmt, ## __VA_ARGS__);
#define aud_error(fmt, ...)	gaim_debug(GAIM_DEBUG_ERROR, "Gaim-Audacious", \
					fmt, ## __VA_ARGS__);

static gint timeout_tag = 0;

GHashTable *stored_status;
GHashTable *stored_userinfo;
GHashTable *stored_alias;

GHashTable *pushed_status;
GHashTable *pushed_userinfo;
GHashTable *pushed_alias;


static void aud_process(gchar *aud_info);

static gboolean
watchdog_func(void)
{
	gint playpos;
	gchar *song = NULL, *tmp = NULL;
	gint session = 0;			/* default session */
	gboolean rv = TRUE;
	size_t dummy;

	if(!xmms_remote_is_playing(0)) {	/* audacious isn't running */
		aud_process(NULL);
		return rv;
	}

	playpos = xmms_remote_get_playlist_pos(session);
	tmp = xmms_remote_get_playlist_title(session, playpos);
	if(botch_utf != NULL)
		song = botch_utf(tmp, strlen(tmp), &dummy);
	else
		song = g_strdup(tmp);

	g_free(tmp);
	tmp = NULL;

	aud_process(song);
	g_free(song);
	song = NULL;

	return rv;
}

static void
aud_process_status(GaimConnection *gc, gchar *aud_info)
{
	gchar *new;
	const gchar *old, *proto;
	GaimAccount *account;
	GaimPresence *presence;
	GaimPlugin *prpl;
	GaimPluginProtocolInfo *prpl_info;
	GaimStatus *status;

	gpointer val; // for hash
	gchar *key;

	account = gaim_connection_get_account(gc);
	presence = gaim_account_get_presence(account);

	proto = gaim_account_get_protocol_id(account);
	prpl = gaim_find_prpl(proto);
	g_return_if_fail(prpl != NULL);

	prpl_info = GAIM_PLUGIN_PROTOCOL_INFO(prpl);
	g_return_if_fail(prpl_info != NULL && prpl_info->set_status != NULL);

	status = gaim_presence_get_active_status(presence);
	g_return_if_fail(status != NULL);

	old = gaim_status_get_attr_string(status, "message");
	if(old == NULL || strlen(old) == 0)
	   return;

	/* generate key for hash table*/
	key = g_strdup_printf("%s %s", account->username, account->protocol_id);

	/* if old contains token, compare with stored_status */
	if(strstr(old, SONG_TOKEN)) {
		g_hash_table_replace(stored_status, g_strdup(key), g_strdup(old));
	}

	/* construct new status message */
	val = g_hash_table_lookup(stored_status, key);
	g_return_if_fail(val != NULL);
	aud_debug("status stored = %s\n", (gchar *)val);

	if(aud_info){
		new = gaim_strreplace(val, SONG_TOKEN, aud_info);
	}
	else {
		new = g_strdup("No song being played.");
	}

	g_return_if_fail(new != NULL);

	/* set status message only if text has been changed */
	val = g_hash_table_lookup(pushed_status, key);
	aud_debug("status pushed = %s\n", (gchar *)val);

	if (!val || g_ascii_strcasecmp(val, new) != 0) {
		g_hash_table_replace(pushed_status, g_strdup(key), g_strdup(new));
		gaim_status_set_attr_string(status, "message", new);
		prpl_info->set_status(account, status);
	}
	g_free(key);
	g_free(new);
}


static void
aud_process_userinfo(GaimConnection *gc, gchar *aud_info)
{
	gchar *new;
	const gchar *old, *proto;
	GaimAccount *account;
	GaimPlugin *prpl;
	GaimPluginProtocolInfo *prpl_info;

	gpointer val; // for hash
	gchar *key;

	account = gaim_connection_get_account(gc);

	proto = gaim_account_get_protocol_id(account);
	prpl = gaim_find_prpl(proto);
	g_return_if_fail(prpl != NULL);

	prpl_info = GAIM_PLUGIN_PROTOCOL_INFO(prpl);
	g_return_if_fail(prpl_info != NULL && prpl_info->set_info != NULL);

	/* retrieve the old user info */
	old = gaim_account_get_user_info(account);		/* it's always from account.xml! */
	if(old == NULL || strlen(old) == 0)
		return;

	/* generate key for hash table*/
	key = g_strdup_printf("%s %s", account->username, account->protocol_id);

	/* if old contains token, compare with stored_status */
	if(strstr(old, SONG_TOKEN)) {
		g_hash_table_replace(stored_userinfo, g_strdup(key), g_strdup(old));
	}

	/* construct new status message */
	val = g_hash_table_lookup(stored_userinfo, key);
	g_return_if_fail(val != NULL);

	aud_debug("userinfo stored = %s\n", (gchar *)val);

	if(aud_info){
		new = gaim_strreplace(val, SONG_TOKEN, aud_info);
	}
	else {
		new = g_strdup("No song being played.");
	}

	g_return_if_fail(new != NULL);

	/* set user info only if text has been changed */
	val = g_hash_table_lookup(pushed_userinfo, key);
	aud_debug("userinfo pushed = %s\n", (gchar *)val);

	if (!val || g_ascii_strcasecmp(val, new) != 0) {
		g_hash_table_replace(pushed_userinfo, g_strdup(key), g_strdup(new));
		prpl_info->set_info(gc, new);
	}
	g_free(key);
	g_free(new);
}

static void
aud_process_alias(GaimConnection *gc, gchar *aud_info)
{
	gchar *new;
	const gchar *old, *proto;
	GaimAccount *account;
	GaimPlugin *prpl;
	GaimPluginProtocolInfo *prpl_info;

	gpointer val; // for hash
	gchar *key;

	account = gaim_connection_get_account(gc);

	proto = gaim_account_get_protocol_id(account);
	prpl = gaim_find_prpl(proto);
	g_return_if_fail(prpl != NULL);

	prpl_info = GAIM_PLUGIN_PROTOCOL_INFO(prpl);
	g_return_if_fail(prpl_info != NULL && prpl_info->alias_buddy != NULL);

	/* retrieve the old alias */
//	old = gaim_url_decode(gaim_account_get_alias(account));
	old = gaim_account_get_alias(account);
	if(old == NULL || strlen(old) == 0) {
		aud_error("couldn't get old alias\n");
		return;
	}
	
	aud_debug("old alias = %s\n", old);

	/* generate key for hash table*/
	key = g_strdup_printf("%s %s", account->username, account->protocol_id);

	aud_debug("alias key = %s\n", key);

	/* if old contains token, compare with stored_status */
	if(strstr(old, SONG_TOKEN)){
		g_hash_table_replace(stored_alias, g_strdup(key), g_strdup(old));
	}

	/* construct new status message */
	val = g_hash_table_lookup(stored_alias, key);
	g_return_if_fail(val != NULL);

	aud_debug("alias stored = %s\n", (gchar *)val);

	if(aud_info){
		new = gaim_strreplace(val, SONG_TOKEN, aud_info);
	}
	else {
		new = gaim_strreplace(val, SONG_TOKEN, "none");
	}

	g_return_if_fail(new != NULL);

	/* set user info only if text has been changed */
	val = g_hash_table_lookup(pushed_alias, key);
	aud_debug("alias pushed = %s\n", (gchar *)val);

	if (!val || g_ascii_strcasecmp(val, new) != 0) {
		g_hash_table_replace(pushed_alias, g_strdup(key), g_strdup(new));
		gaim_account_set_alias(account, new); //oct16
		prpl_info->alias_buddy(gc, account->username, new);
	}
	g_free(key);
	g_free(new);
}

static void
aud_process(gchar *aud_info)
{
	GList *l;
	GaimConnection *gc;

	for (l = gaim_connections_get_all(); l != NULL; l = l->next) {
		gc = (GaimConnection *) l->data;

		/* make sure we're connected */
		if (gaim_connection_get_state(gc) != GAIM_CONNECTED) {
			continue;
		}

		if (gaim_prefs_get_bool(OPT_PROCESS_USERINFO)) {
			aud_process_userinfo(gc, aud_info);
		}

		if (gaim_prefs_get_bool(OPT_PROCESS_STATUS)) {
			aud_process_status(gc, aud_info);
		}

		if (gaim_prefs_get_bool(OPT_PROCESS_ALIAS)) {
			aud_process_alias(gc, aud_info);
		}

	}
}
static void
removekey(gpointer data)
{
	g_free(data);
}

static void
removeval(gpointer data)
{
	g_free(data);
}

static gboolean
restore_alias(GaimConnection *gc, gpointer data)
{
	GaimAccount *account;
	gpointer val = NULL;
	gchar *key = NULL;

	aud_debug("********** restore_alias called **********\n");
	account = gaim_connection_get_account(gc);

	key = g_strdup_printf("%s %s", account->username, account->protocol_id);
	val = g_hash_table_lookup(stored_alias, key);
	g_return_val_if_fail(val != NULL, FALSE);

	aud_debug("write back alias %s\n", val);
	gaim_account_set_alias(account, val); //oct16

	g_free(key);

	return TRUE;
}


static gboolean
load_plugin(GaimPlugin *plugin)
{

	gchar *remote_uri;

	remote_uri = getenv("AUDTOOL_REMOTE_URI");
	audacious_set_session_uri(remote_uri);

	stored_status = g_hash_table_new_full(g_str_hash, g_str_equal, removekey, removeval);
	stored_alias = g_hash_table_new_full(g_str_hash, g_str_equal, removekey, removeval);
	stored_userinfo = g_hash_table_new_full(g_str_hash, g_str_equal, removekey, removeval);

	pushed_status = g_hash_table_new_full(g_str_hash, g_str_equal, removekey, removeval);
	pushed_alias = g_hash_table_new_full(g_str_hash, g_str_equal, removekey, removeval);
	pushed_userinfo = g_hash_table_new_full(g_str_hash, g_str_equal, removekey, removeval);

	timeout_tag = g_timeout_add(15*1000, (gpointer)watchdog_func, NULL);

	/* connect to signing-off signal */
	gaim_signal_connect(gaim_connections_get_handle(), "signing-off", plugin,
						GAIM_CALLBACK(restore_alias), NULL);


	return TRUE;
}

static gboolean
unload_plugin(GaimPlugin *plugin)
{
	aud_debug("gaim-audacious unload called\n");

	g_hash_table_destroy(stored_status);
	g_hash_table_destroy(stored_alias);
	g_hash_table_destroy(stored_userinfo);

	g_hash_table_destroy(pushed_status);
	g_hash_table_destroy(pushed_alias);
	g_hash_table_destroy(pushed_userinfo);

	return TRUE;
}

static GaimPluginPrefFrame *
get_plugin_pref_frame(GaimPlugin *plugin)
{
	GaimPluginPref *pref;
	GaimPluginPrefFrame *frame = gaim_plugin_pref_frame_new();

	/* create gtk elements for the plugin preferences */
	pref = gaim_plugin_pref_new_with_label("Gaim-Audacious Configuration");
	gaim_plugin_pref_frame_add(frame, pref);
	
	pref = gaim_plugin_pref_new_with_name_and_label(OPT_PROCESS_STATUS,
			"Expand " SONG_TOKEN " to song info in the status message");
	gaim_plugin_pref_frame_add(frame, pref);
	
	pref = gaim_plugin_pref_new_with_name_and_label(OPT_PROCESS_USERINFO,
			"Expand " SONG_TOKEN " to song info in the user info");
	gaim_plugin_pref_frame_add(frame, pref);

	pref = gaim_plugin_pref_new_with_name_and_label(OPT_PROCESS_ALIAS,
			"Expand " SONG_TOKEN " to song info in the alias");
	gaim_plugin_pref_frame_add(frame, pref);

	return frame;
}

static GaimPluginUiInfo pref_info =
{
	get_plugin_pref_frame
};

static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_STANDARD,				/**< type	*/
	GAIM_GTK_PLUGIN_TYPE,				/**< ui_req	*/
	0,						/**< flags	*/
	NULL,						/**< deps	*/
	GAIM_PRIORITY_DEFAULT,				/**< priority	*/
	GAIMAUD_PLUGIN_ID,				/**< id		*/
	"Gaim-Audacious",				/**< name	*/
	"1.1.0d5",					/**< version	*/
							/**  summary	*/
	"Automatically updates your Gaim status info	with the currently "
			"playing music in Audacious.",
							/**  desc	*/
	"Automatically updates your Gaim status info with the currently "
			"playing music in Audacious.",
	"Yoshiki Yazawa (yaz@honeyplanet.jp)",		/**< author	*/
	"http://www.honeyplanet.jp",			/**< homepage	*/
	load_plugin,					/**< load	*/
	unload_plugin,					/**< unload	*/
	NULL,						/**< destroy	*/
	NULL,						/**< ui_info	*/
	NULL,						/**< extra_info	*/
	&pref_info,					/**< pref info	*/
	NULL
};

static void
init_plugin(GaimPlugin *plugin)
{
	g_type_init();

	/* add plugin preferences */
	gaim_prefs_add_none(OPT_GAIMAUD);
	gaim_prefs_add_bool(OPT_PROCESS_STATUS, TRUE);
	gaim_prefs_add_bool(OPT_PROCESS_USERINFO, TRUE);
	gaim_prefs_add_bool(OPT_PROCESS_ALIAS, TRUE);
}

GAIM_INIT_PLUGIN(gaim_audacious, init_plugin, info)
